import logging
import subprocess


class SoundManager:
    def __init__(self,
                 player_command: str = "mpg123",
                 argument_format: str = "{}",  # '{}' will be replaced with filename
                 **kwargs):
        self._logger = logging.getLogger(__name__)
        self._player_command = player_command
        self._argument_format = argument_format
        self._filenames = kwargs
        self._logger.debug("Passed kwargs: {}".format(kwargs))
        self._process = None

    def _get_argument(self, filename) -> str:
        return self._argument_format.format(filename)

    def play(self, sound: str):
        self._logger.debug("Playing {}".format(sound))
        self.kill()
        try:
            if sound in self._filenames:
                argument = self._get_argument(self._filenames[sound])
                self._logger.debug("Command is: {}; argument is: {}".format(self._player_command, argument))
                self._process = subprocess.Popen([self._player_command, argument],
                                                 stdin=subprocess.DEVNULL,
                                                 stdout=subprocess.DEVNULL,
                                                 stderr=subprocess.DEVNULL)
            else:
                self._logger.error("Sound {} not found".format(sound))
        except Exception as e:
            self._logger.error("Error playing sound {}: {}".format(sound, e))

    def cleanup(self, timeout=15):
        if self._process and self._process.poll() is None:
            try:
                self._logger.debug("Waiting for sound to finish")
                self._process.wait(timeout)
                self._process.communicate()
            except subprocess.TimeoutExpired:
                self._process.kill()
                self._process.communicate()

    def kill(self):
        if self._process and self._process.poll() is None:
            try:
                self._logger.debug("Killing process")
                self._process.kill()
                self._process.communicate()
            except Exception as e:
                self._logger.error("Exception during kill: {}".format(e))
        else:
            self._logger.debug("Process is None or finished")

# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from ..station import Station, labware_loader, instrument_loader
from ..multi_tube_source import MultiTubeSource
from ..utils import uniform_divide, MoveWithSpeed
import math
import logging

class BioerProtocol(Station):

    def __init__(self,
            debug_mode = False,
            num_samples: int = 96,
            pk_tube_bottom_height = 2,
            mm_tube_bottom_height = 2,
            pcr_bottom_headroom_height = 4.5,
            dw_bottom_height = 13.5,
            mix_bottom_height = 0.5,
            mix_bottom_height_dw = -1.8,
            mm_plate_bottom_height = 1.2,
            dw_elutes_bottom_height = -3,
            pk_volume_tube = 320,
            vol_pk_offset = 5,
            vol_mm_offset = 10,
            elution_volume = 12,
            mm_volume = 17,
            mm_volume_tube = 1800,
            mm_tube_vertical_speed = 25,
            headroom_vol_from_tubes_to_pcr = 60,
            headroom_vol_from_tubes_to_dw = 10,
            control_well_positions = ['G12', 'H12'],
            pause_between_mastermix_and_elutes: bool = True,
            tube_block_model: str = 'opentrons_24_aluminumblock_nest_2ml_screwcap',
            transfer_elutes_phase: bool = False,
            transfer_proteinase_phase: bool = False,
            mix_beads_phase: bool = False,
            mastermix_phase: bool = False,
            vertical_offset = -16,
            slow_vertical_speed: float = 25,
            elution_air_gap = 10,
            final_mix_blow_out_height = -2,
            ** kwargs):

        super(BioerProtocol, self).__init__(
            num_samples = num_samples,
            ** kwargs)

        self._debug_mode = debug_mode
        self._dests_sample = []
        self._max_sample_per_dw = 16
        self._elution_volume = elution_volume
        self._pk_volume = 10
        self._pk_volume_tube = pk_volume_tube
        self._mm_volume = mm_volume
        self._mm_volume_tube = mm_volume_tube
        self._mm_tube_bottom_height = mm_tube_bottom_height
        self._mm_tube_vertical_speed = mm_tube_vertical_speed
        self._dw_bottom_height = dw_bottom_height
        self._mix_bottom_height_dw = mix_bottom_height_dw
        self._pcr_bottom_headroom_height = pcr_bottom_headroom_height
        self._mix_bottom_height = mix_bottom_height
        self._mm_plate_bottom_height = mm_plate_bottom_height
        self._dw_elutes_bottom_height = dw_elutes_bottom_height
        self._pk_tube_source = MultiTubeSource()
        self._mm_tube_source = MultiTubeSource(vertical_speed=25)
        self._control_well_positions = control_well_positions
        self._pk_tube_bottom_height = pk_tube_bottom_height
        self._headroom_vol_from_tubes_to_pcr = headroom_vol_from_tubes_to_pcr
        self._headroom_vol_from_tubes_to_dw = headroom_vol_from_tubes_to_dw
        self._pause_between_mastermix_and_elutes = pause_between_mastermix_and_elutes
        self._tube_block_model = tube_block_model
        self._mix_beads_phase = mix_beads_phase
        self._mastermix_phase = mastermix_phase
        self.transfer_elutes_phase = transfer_elutes_phase
        self.transfer_proteinase_phase = transfer_proteinase_phase
        self._vertical_offset = vertical_offset
        self._vol_pk_offset = vol_pk_offset
        self._vol_mm_offset = vol_mm_offset
        self._slow_vertical_speed = slow_vertical_speed
        self._elution_air_gap = elution_air_gap
        self._final_mix_blow_out_height = final_mix_blow_out_height
        self._s300_fake_aspirate: bool = True

        if self._num_samples > 80:
            self._dws = ['8', '9', '5', '6', '2', '3']
        elif self._num_samples > 64:
            self._dws = ['8', '9', '5', '6', '2']
        elif self._num_samples > 48:
            self._dws = ['8', '9', '5', '6']
        elif self._num_samples > 32:
            self._dws = ['8', '9', '5']
        elif self._num_samples > 16:
            self._dws = ['8', '9']
        else:
            self._dws = ['8']

    @labware_loader(1, "_tips200")
    def load_tips200(self):
        self._tips200 = [self._ctx.load_labware('opentrons_96_filtertiprack_200ul', slot, '200µl filter tiprack multi')
                                for slot in ['7', '10']]

    @labware_loader(2, "_tips200_single")
    def load_tips200_single(self):
        self._tips200_single = [self._ctx.load_labware('opentrons_96_tiprack_300ul', '11', '200µl filter tiprack single')]

    @labware_loader(3, "_plate_elute")
    def load_dest_plate_elute(self):
        self._dest_plate_elute = self._ctx.load_labware('opentrons_96_aluminumblock_nest_wellplate_100ul', '1', 'chilled elution plate')

    @labware_loader(4, "_tube_block")
    def load_tube_block(self):
        self._tube_block = self._ctx.load_labware(self._tube_block_model, '4', 'screw tube aluminum block for proteinase and mastermix')

    @labware_loader(5, "_dests_plate")
    def load_dests_plate(self):
        self._dests_plates = [self._ctx.load_labware('nest_96_wellplate_2ml_deep', slot, '96-deepwell sample plate' + str(e + 1))
                                for e, slot in enumerate(self._dws)]

    @instrument_loader(0, "_p300")
    def load_p300(self):
        self._p300 = self._ctx.load_instrument('p300_multi_gen2', 'left', tip_racks=self._tips200)

    @instrument_loader(0, "_s300")
    def load_s300(self):
        self._s300 = self._ctx.load_instrument('p300_single_gen2', 'right', tip_racks=self._tips200_single)

    def _tipracks(self) -> dict:
        return {"_tips200":"_p300", "_tips200_single":"_s300"}


    @property
    def sample_dests_wells(self):
        return self._dest_plate_elute.wells()[:self._num_samples]


    def is_well_in_samples(self, well):
        """
        Function that check if a well is within the samples well.
        :param well: well to check
        :return: True if the well is included in the samples list.
        """
        return well in self.sample_dests_wells


    @property
    def control_dests_wells(self):
        return [self._dest_plate_elute.wells_by_name()[i] for i in self._control_well_positions]


    @property
    def control_wells_not_in_samples(self):
        """
        :return: a list of wells for controls that are not already filled with the 8-channel pipette
        """
        return [c for c in self.control_dests_wells if not self.is_well_in_samples(c)]


    @property
    def destinations(self):
        for plate in self._dests_plates:
            self._dests_sample += plate.wells()[:8]
            self._dests_sample += plate.wells()[48:56]
        return self._dests_plates


    @property
    def proteinase_destinations(self):
        return [row for plate in self._dests_plates for row in plate.columns()[::6]]


    @property
    def proteinase_destinations_unique(self):
        dests_pk_unique = []
        for d in self.proteinase_destinations:
            dests_pk_unique += d
        return dests_pk_unique

    @property
    def beads_destinations(self):
        return [row for plate in self._dests_plates for row in plate.rows()[0]][5::6]


    def transfer_proteinase(self):
        done_samples = 0
        num_cycle = 1
        num_samples_per_fill = 16

        self.logger.info("Trasferring proteinase from tube to deepwells")

        self.pick_up(self._s300)
        while done_samples < self._num_samples:
            samples_to_do = self.proteinase_destinations_unique[done_samples:(done_samples + num_samples_per_fill)]
            # self.logger.info("Cycle {} - samples to do: {}".format(num_cycle, samples_to_do))
            # self.logger.info(
            #     "PK:Cycle {} - before filling: samples done: {}, samples to do: {}".format(num_cycle, done_samples,
            #                                                                                len(samples_to_do)))
            if num_cycle > 1:
                self._vol_pk_offset = 0
            vol_pk = len(samples_to_do) * self._pk_volume
            if num_cycle == 1:
                self._s300.aspirate(self._vol_pk_offset, self._tube_block.wells()[0].bottom(self._pk_tube_bottom_height))
            self._pk_tube_source.prepare_aspiration(vol_pk, self._pk_tube_bottom_height)
            self._pk_tube_source.aspirate(self._s300)
            #self.logger.info("Aspirating {} at: {} mm".format(vol_pk, self._pk_tube_bottom_height))
            for s, ss in enumerate(samples_to_do):
                self._s300.dispense(self._pk_volume, ss.bottom(self._dw_bottom_height))
                self._s300.touch_tip(ss, 0.6, self._vertical_offset)
            done_samples += num_samples_per_fill
            # self.logger.info("PK:Cycle {} - after distribution: samples done: {}".format(num_cycle, done_samples))
            num_cycle += 1
        if self._s300.has_tip:
            self.drop(self._s300)


    def mix_beads(self):
        done_mix = 0
        destbeads = self.beads_destinations[done_mix:(done_mix + len(self.beads_destinations))]

        self.logger.info("Mix beads in deepwells")

        self.pick_up(self._p300)
        for b, d in enumerate(destbeads):
            self._p300.mix(3, 15, d.bottom(self._mix_bottom_height_dw))
            done_mix = done_mix + len(self.beads_destinations)
        if self._p300.has_tip:
            self.drop(self._p300)


    def transfer_mastermix(self):
        done_samples = 0
        num_cycle = 1
        num_samples_per_fill = 8

        samples_with_controls = self.sample_dests_wells
        samples_with_controls += self.control_wells_not_in_samples

        self.logger.info("Trasferring mastermix from tube to pcr plate")

        self.pick_up(self._s300)

        while done_samples < self._num_samples + len(self.control_wells_not_in_samples):
            samples_to_do = samples_with_controls[done_samples:(done_samples + num_samples_per_fill)]
            # self.logger.info("Cycle {} - samples to do: {}".format(num_cycle, samples_to_do))
            # self.logger.info("MM:Cycle {} - before filling: samples done: {}, samples to do: {}".format(num_cycle,
            #                                                                     done_samples, len(samples_to_do)))

            if self._s300_fake_aspirate:
                self._s300_fake_aspirate = False
                self._s300.aspirate(10, self._tube_block.wells()[0].top())
                self._s300.dispense(10)
            if num_cycle > 1:
                self._vol_mm_offset = 0
            vol_mm = (len(samples_to_do) * self._mm_volume) + self._vol_mm_offset
            self._mm_tube_source.prepare_aspiration(vol_mm, min_height=self._mm_tube_bottom_height)
            self._mm_tube_source.aspirate(self._s300)
            # self.logger.info("Aspirating at: {} mm".format(self._mm_tube_bottom_height))
            for s, ss in enumerate(samples_to_do):
                with MoveWithSpeed(self._s300,
                                   from_point=ss.bottom(self._mm_plate_bottom_height + 5),
                                   to_point=ss.bottom(self._mm_plate_bottom_height),
                                   speed=self._slow_vertical_speed, move_close=False):
                    self._s300.dispense(self._mm_volume)
            done_samples += num_samples_per_fill
            # self.logger.info("MM:Cycle {} - after distribution: samples done: {}".format(num_cycle, done_samples))
            num_cycle += 1
        if self._s300.has_tip:
            self.drop(self._s300)


    def transfer_elutes(self):
        set_of_source = math.ceil(self._num_samples / self._max_sample_per_dw)
        source_plate = [row for plate in self._dests_plates for row in plate.columns()[4::6]]
        dests_sample_elute = self._dest_plate_elute.columns()[:self.num_cols]

        if self._pause_between_mastermix_and_elutes:
            self.dual_pause(self.get_msg_format("insert deepwell", ", ".join([str(d) for d in self._dests_plates])))

        self.logger.info("Trasferring elutes from deepwells to pcr plate")

        done_col = 0
        for e in range(set_of_source):
            source_el = source_plate[done_col:(done_col + len(source_plate))]
            for s, d in zip(source_el, dests_sample_elute):
                samples = []
                #self.logger.info("s,d:{}\n{}".format(s, d))
                if any([w in d for w in self.control_dests_wells]):
                    # self.logger.info("Using single pipette")
                    for i, (w, z) in enumerate(zip(s, d)):
                        if z not in self.control_dests_wells:
                            samples.append((w, z))
                    pipette = self._s300
                else:
                    samples.append((s[0], d[0]))
                    pipette = self._p300

                for t, o in samples:
                    #self.logger.info("samples:{}".format(samples))
                    self.pick_up(pipette)
                    pipette.aspirate(self._elution_volume, t.bottom(self._dw_elutes_bottom_height))
                    pipette.air_gap(self._elution_air_gap)
                    pipette.dispense(self._elution_air_gap, o.top())
                    pipette.dispense(self._elution_volume, o.bottom(self._pcr_bottom_headroom_height))
                    pipette.mix(5, 20, o.bottom(self._mix_bottom_height))
                    pipette.blow_out(o.top(self._final_mix_blow_out_height))
                    pipette.air_gap(self._elution_air_gap)
                    self.drop(pipette)
                done_col = done_col + len(source_plate)

    def body(self):

        num_dw = math.ceil(self._num_samples / self._max_sample_per_dw)

        self.logger.info("Controls in {}".format(self._control_well_positions))

        #proteinase_destinations
        dests_pk_unique = []
        for d in self.proteinase_destinations:
            dests_pk_unique += d
        #self.logger.info("dests_pk_unique: {}".format(dests_pk_unique))

        #tube_block_pk
        volume_for_samples = self._pk_volume * self._num_samples
        volume_to_distribute_to_dw = volume_for_samples
        num_pk_tube = math.ceil(volume_for_samples / self._pk_volume_tube)

        [self._pk_tube_source.append_tube_with_vol(t, self._pk_volume_tube) for t in self._tube_block.wells()[0:num_pk_tube]]
        self.logger.info("We need {} tubes with {}ul of proteinase each in {}".format(num_pk_tube,
                                                                                      self._pk_volume_tube,
                                                                                      self._pk_tube_source.locations_str))

        pk_requirements = self.get_msg_format("load pk tubes",
                                           num_pk_tube,
                                           self._pk_volume_tube,
                                           self._pk_tube_source.locations_str)


        #tube_block_mm
        volume_for_controls = len(self.control_wells_not_in_samples) * self._mm_volume
        volume_for_samples = self._mm_volume * self._num_samples
        volume_to_distribute_to_pcr_plate = volume_for_samples + volume_for_controls
        num_tubes, vol_per_tube = uniform_divide(
            volume_to_distribute_to_pcr_plate + self._headroom_vol_from_tubes_to_pcr, self._mm_volume_tube)
        mm_tubes = self._tube_block.wells()[:num_tubes]
        available_volume = volume_to_distribute_to_pcr_plate / len(mm_tubes)
        if self._headroom_vol_from_tubes_to_pcr > 0:
            if self._vol_mm_offset < (self._headroom_vol_from_tubes_to_pcr / len(mm_tubes)):
                available_volume += self._vol_mm_offset
        assert vol_per_tube > available_volume, \
            "Error in volume calculations: requested {}ul while total in tubes {}ul".format(available_volume, vol_per_tube)

        [self._mm_tube_source.append_tube_with_vol(t, available_volume) for t in self._tube_block.wells()[len(self._tube_block.wells())-num_tubes::]]

        self.logger.info("We need {} tubes with {}ul of mastermix each in {}".format(num_tubes,
                                                                                      vol_per_tube,
                                                                                      self._mm_tube_source.locations_str))
        mmix_requirements = self.get_msg_format("load mm tubes",
                                           num_tubes,
                                           vol_per_tube,
                                           self._mm_tube_source.locations_str)

        control_positions = self.get_msg_format("control positions",
                                                self._control_well_positions)

        #transfer_proteinase
        if self.transfer_proteinase_phase:
            self.pause(pk_requirements, home=False)
            self.stage = "Transfer proteinase"
            self.transfer_proteinase()

        #mix_beads
        if self._mix_beads_phase:
            self.stage = "Mix beads"
            self.mix_beads()

        #Bioer phase
        if (self.transfer_proteinase_phase or self._mix_beads_phase) and (self._mastermix_phase or self.transfer_elutes_phase):
            self.dual_pause("Move deepwells to Bioer")

        #transfer mastermix
        if self._mastermix_phase:
            self.pause(control_positions, home=False)
            self.pause(mmix_requirements, home=False)
            self.stage = "Transfer mastermix"
            self.transfer_mastermix()

        #transfer elutes
        if self.transfer_elutes_phase:
            self.stage = "Transfer elutes"
            self.transfer_elutes()

    def drop(self, pip):
        if self._debug_mode:
            pip.return_tip()
        else:
            super(BioerProtocol, self).drop(pip)

# class BioerPreparationToBioer(BioerProtocol):
#     def __init__(self, ** kwargs):
#         super(BioerPreparationToBioer, self).__init__(
#             mix_beads_phase = False,
#             transfer_proteinase_phase = True,
#             ** kwargs)

class BioerPreparationToPcr(BioerProtocol):
    def __init__(self, ** kwargs):
        super(BioerPreparationToPcr, self).__init__(
            mastermix_phase = True,
            transfer_elutes_phase = True,
            ** kwargs)


class BioerPreparationToPcrTechogenetics(BioerPreparationToPcr):
    def __init__(self, ** kwargs):
        super(BioerPreparationToPcrTechogenetics, self).__init__(
            mm_volume = 20,
            elution_volume = 20,
            mm_volume_tube = 1300,
            **kwargs)

# protocol for loading in Opentrons App or opentrons_simulate
# =====================================
logging.getLogger(BioerProtocol.__name__).setLevel(logging.INFO)
metadata = {'apiLevel': '2.7'}
station = BioerProtocol(num_samples = 32)

def run(ctx):
    return station.run(ctx)

if __name__ == "__main__":
    BioerProtocol(metadata={'apiLevel':'2.7'}).simulate()




import json
import os
import time

from opentrons.protocol_api import ProtocolContext, InstrumentContext, PairedInstrumentContext
from opentrons.protocol_api.labware import Well
from opentrons.types import Location
import logging
import math
import requests
from itertools import tee, cycle, islice, chain, repeat
from typing import Tuple, Union, Iterable, Callable, Optional, Dict, Any


class ProtocolContextLoggingHandler(logging.Handler):
    """Logging Handler that emits logs through the ProtocolContext comment method"""
    def __init__(self, ctx: ProtocolContext, *args, **kwargs):
        super(ProtocolContextLoggingHandler, self).__init__(*args, **kwargs)
        self._ctx = ctx
    
    def emit(self, record):
        try:
            self._ctx.comment(self.format(record))
        except Exception:
            self.handleError(record)


class LocalWebServerLogger:
    def __init__(self, ip: Optional[str] = None, endpoint: str = ":5002/log", timeout: float = 2,
                 errorlogger: logging.Logger = logging.getLogger("LWSLog"), *args, **kwargs):
        super(LocalWebServerLogger, self).__init__(*args, **kwargs)
        self.ip = ip
        self.endpoint = endpoint
        self.timeout = timeout
        self.level = 0
        self.last_dollar = None
        self.logfail_count = dict()  # dictionary of url, Number of exceptions during log
        self.logfail_max = 5       # Retries sending log, after that disable send.
        self.errorlogger = errorlogger
    
    @property
    def url(self) -> str:
        return "http://{}{}".format(self.ip, self.endpoint)
    
    def format(self, record):
        if self.last_dollar == record['$']:
            if record['$'] == 'before':
                self.level += 1
            else:
                self.level -= 1
        self.last_dollar = record['$']
        if record['$'] == 'before':
            return ' '.join(['\t' * self.level, record['payload'].get('text', '').format(**record['payload'])])
    
    def __call__(self, record: Dict[str, Any]):
        s = self.format(record)
        url = self.url
        if url and s:
            self.logfail_count.setdefault(url, 0)
            if self.logfail_count[url] < self.logfail_max:
                try:
                    requests.post(
                        url,
                        s.encode('utf-8'),
                        headers={'Content-type': 'text/plain; charset=utf-8'},
                        timeout=self.timeout,
                    )
                    self.logfail_count[url] = 0
                except requests.exceptions.ConnectTimeout as e:
                    self.logfail_count[url] += 1
                    self.errorlogger.error("Exception during lws log {} of {}: {}".format(self.logfail_count[url],
                                                                                            self.logfail_max, e))
                    if self.logfail_count[url] == self.logfail_max:
                        self.errorlogger.error("Disabling lws log, max retries encountered for url {}".format(url))
                except Exception:
                    pass
            else:
                self.errorlogger.debug("Skipping lws log, max retries encountered for url {}".format(url))


def mix_bottom_top(pip,
                   reps: int,
                   vol: float,
                   pos: Callable[[float], Location],
                   bottom: float,
                   top: float,
                   last_dispense_rate: float = None):
    """Custom mixing procedure aspirating at the bottom and dispensing at the top
    :param pip: The pipette
    :param reps: Number of repeats
    :param vol: Volume to mix
    :param pos: Method for getting the position
    :param bottom: Offset for the bottom position
    :param top: Offset for the top position
    :param last_dispense_rate: Dispense rate for the last transfer to keep the tip clean"""

    for i in range(reps):
        if i+1 == reps and last_dispense_rate is not None:
            pip.flow_rate.dispense = last_dispense_rate
        pip.aspirate(vol, pos(bottom))
        pip.dispense(vol, pos(top))


def mix_walk(
    pip,
    reps: int,
    vol: float,
    aspirate_locs: Union[Iterable, Location],
    dispense_locs: Optional[Union[Iterable, Location]] = None,
    speed: Optional[float] = None,
    logger: Optional[logging.getLoggerClass()] = None
):
    """Custom mixing procedure aspirating and dispensing at custom positions
    :param pip: The pipette
    :param reps: Number of repeats
    :param vol: Volume to mix
    :param aspirate_locs: Position(s) at which to aspirate. If less than reps, they are cycled over.
    :param dispense_locs: Position(s) at which to dispense (optional). If less than reps, they are cycled over. If not specified, dispense in place
    :param speed: Speed for moving the pipette around the mixing position (optional). At the end, the previous speed is restored
    :param logger: Logger for debugging information (optional)"""
    if isinstance(aspirate_locs, Location):
        aspirate_locs = [aspirate_locs]
    if dispense_locs is None:
        aspirate_locs, dispense_locs = tee(aspirate_locs, 2)
    elif isinstance(dispense_locs, Location):
        dispense_locs = [dispense_locs]
    
    old_speed = pip.default_speed
    
    for b, a, d in islice(zip(chain([True], repeat(False)), cycle(aspirate_locs), cycle(dispense_locs)), reps):
        if b and speed is not None:
            pip.move_to(a)
            pip.default_speed = speed
            if logger is not None:
                logger.debug('set speed to {}'.format(speed))
        if logger is not None:
            logger.debug('mixing at {} and {}'.format(a, d))
        pip.aspirate(vol, a)
        pip.dispense(vol, d)
    
    if logger is not None and speed is not None:
        logger.debug('set speed to {}'.format(old_speed))
    pip.default_speed = old_speed


def uniform_divide(total: float, mpp: float) -> Tuple[int, float]:
    """Return the minimum number of partitions and the quantity per partition that uniformly divide a given quantity
    :param total: The total quantity to divide
    :param mpp: Maximum quantity per partition
    :returns: The minimum number of partitions and the quantity in each partition"""
    n = int(math.ceil(total / mpp))
    p = total / n if n else 0
    return n, p


class WellWithVolume:
    """Class to make easy to calcualate and use the height of the liquid in a well"""
    def __init__(self, well: Well, initial_vol: float, min_height: float = 0.5, headroom_height: float = 2.0):
        """Class initialization
        :param total_vol: the total volume expected in the well
        :param min_height: optional, the minimum height
        :param headroom_height: optional, the height in mm of liquid above the tip expected after the aspirate
        """
        self._well = well

        assert initial_vol >= 0, "WellWithVolume initial volume negative for well {}".format(well)
        self._volume = initial_vol

        self._min_height = min_height
        self._headroom_height = headroom_height

    @property
    def height(self) -> float:
        if self._well.diameter:
            remaining_height = self._volume / (math.pi * (self._well.diameter/2)**2)
        else:
            remaining_height = self._volume / self._well.length**2

        remaining_height -= self._headroom_height
        final_height = max(remaining_height, self._min_height)
        return final_height

    def extract_vol_and_get_height(self, aspirate_vol: float) -> float:
        """Return the maximum height to aspirate safely a volume from a well
        :param aspirate_vol: The volume to aspirate
        :returns: the maximum height to aspirate safely the volume in the well"""
        self._volume = self._volume - aspirate_vol if self._volume > aspirate_vol else 0

        return self.height

    def fill(self, vol: float):
        """Add the volume passed to the well"""

        self._volume = self._volume + vol
        if self._volume < 0:
            self._volume = 0


class MoveWithSpeed:
    """Class to make easy aspirate and dispense approaching with a defined speed.
    This should be useful with viscous liquid in order not to break the external meniscus on tip avoiding drops"""
    def __init__(self, pip: Union[InstrumentContext, PairedInstrumentContext],
                 from_point: Location,
                 to_point: Location,
                 speed,
                 move_close: bool = True,
                 go_away: bool = True,
                 logger: logging.Logger = logging.getLogger("MovieWithSpeed")):
        """Class initialization
        :param pip: pipette to move;
        :param from_point: first point to reach at default speed and last point to leave;
        :param to_point: target point; from from_point to this point the speed parameter is used;
        :param speed: speed for the movement between from_point and to_point and back
        :param move_close: speed is set going towards to_point
        :param go_away: speed is set going back to from_point."""
        self._move_close = move_close
        self._go_away = go_away
        self._pip = pip
        self._from_point = from_point
        self._to_point = to_point
        self._speed = speed
        self._logger = logger

        # We want to force_direct only if from and to point are on the same well.
        if self._to_point.labware.is_well and self._from_point.labware.is_well:
            self._logger.debug("Both labware are well")
            self._force_direct = self._to_point.labware.as_well() == self._from_point.labware.as_well()
        else:
            self._logger.debug("One labware is not well {}, {}".format(self._from_point.labware, self._to_point.labware))
            self._force_direct = False
        self._logger.debug("Force direct is: {}".format(self._force_direct))

    def __enter__(self):
        self._logger.debug("Moving close. Force direct is: {}".format(self._force_direct))
        if self._move_close:
            self._pip.move_to(self._from_point)
            self._pip.move_to(self._to_point, force_direct=self._force_direct, speed=self._speed)
        else:
            self._pip.move_to(self._to_point)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._logger.debug("Moving close. Force direct is: {}".format(self._force_direct))
        self._pip.move_to(self._to_point)       # for safety, since we've a force_direct set on next move_to
        self._pip.move_to(self._from_point, force_direct=self._force_direct, speed=self._speed if self._go_away else None)


def get_labware_json_from_filename(filename: str = ""):
    with open(os.path.join(os.path.dirname(__file__), 'labware', filename)) as f:
        return json.load(f)


class DelayManager:
    def __init__(self,
                 logger: logging.Logger = logging.getLogger("DelayManager")):
        self._logger = logger
        self._start_time = 0
        self._duration = 0

    def start(self):
        self._start_time = time.time()
        self._duration = 0
        self._logger.info("Entering; now is {}".format(time.strftime("%H:%M", time.localtime(self._start_time))))

    def stop(self):
        self._duration = (time.time() - self._start_time) / 60
        self._logger.info("Exiting; actual duration is {:.2f} minutes".format(self._duration))

    def get_remaining_delay(self, seconds: float = 0, minutes: float = 0) -> float:
        total_time_minutes = seconds/60 + minutes
        delay = (total_time_minutes - self._duration) if total_time_minutes > self._duration else 0
        self._logger.info("Returning delay of {:.2f} minutes".format(delay))
        return delay

    @property
    def start_time(self):
        return self._start_time

# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from .hardware import HardwareHelper
from opentrons.types import Mount, Point
from opentrons.hardware_control.types import PipettePair

import logging

class MovementManager:
    def __init__(self, ctx, logger=logging.getLogger(__name__)):
        self._ctx = ctx
        self._hw_helper = HardwareHelper(ctx)
        self._home_pos: dict = {}
        self._logger = logger

    def move_to_home(self, safety_margin=10, force: bool = False):
        ''' Move pipettes near home
            without really homing to save time since homing is not needed
            :param safety_margin: the distance to keep from home position (not to hit the home switches)
            :param force: force the movement without checking if the gantry is near home
        '''
        self._ctx.comment("Moving near home")
        if Mount.LEFT not in self._home_pos or Mount.RIGHT not in self._home_pos:
            self._ctx.home()
            self.save_home_pos_if_needed()
        else:
            actual_pos = dict()
            for m in [Mount.LEFT, Mount.RIGHT]:
                actual_pos[m] = self._hw_helper.get_gantry_position(m)
                self._logger.debug("Got ActualPos {} for mount {}".format(actual_pos[m], m))
                if actual_pos[m].z != self._home_pos[m].z:
                    self._logger.debug("Retracting {}".format(m))
                    self.retract()
                else:
                    self._logger.debug("Not retracting {}".format(m))

            self._logger.debug("actual pos L: {}".format(actual_pos[Mount.LEFT]))
            self._logger.debug("actual pos R: {}".format(actual_pos[Mount.RIGHT]))
            self._logger.debug("required pos L: {}".format(self._home_pos[Mount.LEFT]))
            self._logger.debug("required pos R: {}".format(self._home_pos[Mount.RIGHT]))

            if force or self._need_home(actual_pos, safety_margin):
                target_mount = Mount.RIGHT if self._mount is Mount.RIGHT else Mount.LEFT     # avoiding PairedPipette Mounts
                target_pos = self._home_pos[target_mount]
                delta_pos = Point(target_pos.x - actual_pos[target_mount].x - safety_margin,
                                  target_pos.y - actual_pos[target_mount].y - safety_margin, 0)
                self._hw_helper.move_rel(target_mount, delta_pos)

    def home(self):
        self._ctx.home()
        self.save_home_pos_if_needed()

    def save_home_pos_if_needed(self):
        ''' saves the home position for the first time
            ==>> WARNING: To be called right AFTER a home command <<==
        '''
        for m in [Mount.LEFT, Mount.RIGHT]:
            self._logger.debug("Saving mount {}".format(m))
            if self._home_pos.get(m, None) is None:
                self._home_pos[m] = self._hw_helper.get_gantry_position(m, True)

    def retract(self):
        ''' We try to retract the pipette attached
            if we have zero or two pipette retract both two axes.'''
        self._logger.debug("Retracting mount {}".format(self._mount))
        self._hw_helper.retract(self._mount)

    @property
    def _mount(self):
        ''' Get the actual mount to use for pseudo-homing
            if only one pipette is present that mount is taken, otherwise the left one '''
        if len(self._ctx.loaded_instruments) == 1:
            mount = Mount.LEFT if 'left' in self._ctx.loaded_instruments else Mount.RIGHT
        else:
            mount = PipettePair.PRIMARY_LEFT
        self._logger.debug("Returning mount {}".format(mount))
        return mount

    def _need_home(self, actual_positions, safety_margin: float) -> bool:
        x_is_near_home = False
        y_is_near_home = False

        for m in [Mount.LEFT, Mount.RIGHT]:
            if m in actual_positions:
                self._logger.debug("Mount {}, Actual x: {}, trash x: {}".format(m, actual_positions[m].x, self._ctx.fixed_trash['A1'].center().point.x))
                self._logger.debug("Mount {}, Actual y: {}, trash y: {}".format(m, actual_positions[m].y,
                                                                               self._ctx.fixed_trash[
                                                                                   'A1'].center().point.y))
                x_is_near_home = x_is_near_home or \
                            actual_positions[m].x + safety_margin >= self._ctx.fixed_trash['A1'].center().point.x or \
                            actual_positions[m].x + safety_margin == self._home_pos[m].x
                y_is_near_home = y_is_near_home or \
                            actual_positions[m].y + safety_margin >= self._ctx.fixed_trash['A1'].center().point.y or \
                            actual_positions[m].y + safety_margin == self._home_pos[m].y
        self._logger.debug("X near: {}, Y near: {}".format(x_is_near_home, y_is_near_home))
        return not (x_is_near_home and y_is_near_home)




import json

from ..station import instrument_loader, labware_loader
from ..utils import uniform_divide, WellWithVolume, MoveWithSpeed, mix_bottom_top
from .technogenetics import StationBTechnogenetics
from ..paired_pipette import PairedPipette
from opentrons.types import Point
from itertools import repeat
import os


class StationBTechnogeneticsPairedPipette(StationBTechnogenetics):

    def __init__(self,
                 drop_height=-10,
                 supernatant_removal_side=1.5,
                 supernatant_removal_side_last_transfer=0.5,
                 pick_up_single=True,
                 **kwargs):
        """ Build a :py:class:`.StationBTechnogeneticsPairedPipette`.
        :param pick_up_single: whether or not to force a single-pipette pick up even when paired operation
                               is needed. This should mitigate the pulled-up tiprack problem seen expecially
                               with paired pipette.
        """
        super(StationBTechnogeneticsPairedPipette, self).__init__(
            drop_height=drop_height,
            supernatant_removal_side=supernatant_removal_side,
            supernatant_removal_side_last_transfer=supernatant_removal_side_last_transfer,
            **kwargs)
        self._pick_up_single = pick_up_single       # Force every paired pickup to be done with one pipette at a time.

    @property
    def _res12_labware_def(self):
        # re-loading labware to avoid bug on reagent reservoir (see https://github.com/Opentrons/opentrons/issues/7793)
        reservoir_file = os.path.join(os.path.split(__file__)[0], "nest_12_reservoir_15ml_modified.json")
        self.logger.info("Loading reservoir labware from: {}".format(reservoir_file))
        with open(reservoir_file) as labware_file:
            labware_def = json.load(labware_file)
        return labware_def

    @labware_loader(8, "_res12")
    def load_res12(self):
        self._res12 = self._ctx.load_labware_from_definition(self._res12_labware_def, 5, 'Trough with WashReagent A')

    @labware_loader(9, "_elut12")
    def load_elut12(self):
        self._elut12 = self._ctx.load_labware_from_definition(self._res12_labware_def, 2, 'Trough with Wash B and Elution buffer')

    @property
    def wash1(self):
        return self._res12.rows()[0][:6]

    @property
    def wash2(self):
        return self._elut12.rows()[0][:6]

    @property
    def water(self):
        return self._elut12.rows()[0][11]

    @instrument_loader(0, "_m300r")
    def load_m300_right(self):
        self._m300r = self._ctx.load_instrument('p300_multi_gen2', 'right', tip_racks=self._tips300)
        self.logger.info("Loaded right pipette")
        if self._bind_aspiration_rate:
            self._m300r.flow_rate.aspirate = self._bind_aspiration_rate
        if self._bind_dispense_rate:
            self._m300r.flow_rate.dispense = self._bind_dispense_rate
        if self._bind_blowout_rate:
            self._m300r.flow_rate.blow_out = self._bind_blowout_rate

    def body(self):
        PairedPipette.setup(self._m300, self._m300r, self, pick_up_single=self._pick_up_single)
        super(StationBTechnogeneticsPairedPipette, self).body()

    def mix_samples(self):
        well_with_volume = WellWithVolume(self.temp_samples_m[0],
                                          initial_vol=self._starting_vol - self._sample_mix_vol,
                                          min_height=self._sample_mix_height,
                                          headroom_height=0)
        with PairedPipette(self._tempplate, self.temp_samples_m, start_at="mix sample") as tp:
            tp.set_flow_rate(aspirate=self._mix_samples_rate, dispense=self._mix_samples_rate)
            tp.pick_up()
            # Custom mix_bottom_top
            for i in range(self._sample_mix_times):
                if i + 1 == self._sample_mix_times and self._mix_samples_last_rate is not None:
                    tp.set_flow_rate(dispense=self._mix_samples_last_rate)
                tp.aspirate(self._sample_mix_vol,
                            locationFrom="target",
                            well_modifier="bottom({})".format(self._sample_mix_height))
                tp.dispense(self._sample_mix_vol,
                            locationFrom="target",
                            well_modifier="bottom({})".format(well_with_volume.height))
            tp.move_to(locationFrom="target", well_modifier="top(0)", speed=self._sample_vertical_speed)
            tp.air_gap(self._bind_air_gap)
            tp.drop_tip()

    def remove_supernatant(self, vol: float, stage: str = "remove supernatant"):
        self._ctx.comment("Supernatant side: {}, {}".format(self._supernatant_removal_side,
                                                            self._supernatant_removal_side_last_transfer))

        num_trans, vol_per_trans = uniform_divide(vol-self._vol_last_trans,
                                                  self._pipette_max_volume - self._supernatant_removal_air_gap)

        waste_locs = list(repeat(self._waste, len(self.mag_samples_m)))

        sides = []
        sides_last_transfer = []
        for i, m in enumerate(self.mag_samples_m):
            side = -1 if i % 2 == 0 else 1
            sides.append(Point(x=side * self._supernatant_removal_side))
            sides_last_transfer.append(Point(x=side * self._supernatant_removal_side_last_transfer))

        with PairedPipette(self._magplate, self.mag_samples_m, waste_locs=waste_locs, start_at=stage) as tp:
            well_with_volume = WellWithVolume(self.mag_samples_m[0], vol, min_height=self._h_bottom)
            tp.set_flow_rate(aspirate=self._supernatant_removal_aspiration_rate_first_phase,
                             dispense=self._supernatant_removal_dispense_rate)
            tp.pick_up()
            for i in range(num_trans):
                # tp.move_to()  # we want center() but for now it is not in available commands
                height = well_with_volume.extract_vol_and_get_height(vol_per_trans)

                if i > 0:
                    tp.dispense(self._supernatant_removal_air_gap, locationFrom="target", well_modifier="top(0)")
                tp.aspirate(vol_per_trans,
                            locationFrom="target",
                            well_modifier="bottom({})".format(height),
                            well_move_points=sides)
                tp.air_gap(self._supernatant_removal_air_gap)
                tp.dispense(vol_per_trans+self._supernatant_removal_air_gap, locationFrom="waste_locs")
                tp.air_gap(self._supernatant_removal_air_gap)

            tp.comment("Supernatant removal: last transfer in {} step".format(self._n_bottom))
            tp.dispense(self._supernatant_removal_air_gap, locationFrom="target", well_modifier="top(0)")

            tp.set_flow_rate(aspirate=self._supernatant_removal_aspiration_rate)
            for i in range(self._n_bottom):
                aspirate_height = self._h_bottom - i * (self._h_bottom / (self._n_bottom - 1))
                tp.comment("Aspirating at {}".format(aspirate_height))
                tp.aspirate(self._supernatant_removal_last_transfer_max_vol/self._n_bottom,
                            locationFrom="target",
                            well_modifier="bottom({})".format(aspirate_height),
                            well_move_points=sides_last_transfer)

            back_step = 0.1
            n_back_step = self._n_bottom

            for _ in range(n_back_step):
                aspirate_height = aspirate_height + back_step
                tp.comment("Moving up at {}".format(aspirate_height))
                tp.move_to(locationFrom="target",
                           well_modifier="bottom({})".format(aspirate_height),
                           well_move_points=sides_last_transfer)

            tp.air_gap(self._supernatant_removal_air_gap)
            tp.dispense(self._supernatant_removal_last_transfer_max_vol + self._supernatant_removal_air_gap,
                        locationFrom="waste_locs")
            tp.air_gap(self._supernatant_removal_air_gap)
            tp.drop_tip()

    def elute(self, positions=None, transfer: bool = False, stage: str = "elute"):
        if positions is None:
            positions = self.temp_samples_m
        self._magdeck.disengage()
        self._elute(positions=positions, transfer=transfer, stage=stage)
        self._magdeck.disengage()

    def _elute(self, positions=None, transfer: bool = True, stage: str = "elute"):
        """Resuspend beads in elution"""
        # if positions is None:
        #     positions = self.mag_samples_m
        side_locs = []
        for i, m in enumerate(positions):
            side = 1 if i % 2 == 0 else -1
            side_locs.append(m.bottom(self._bottom_headroom_height).move(Point(x=side * 2)))

        with PairedPipette(self._tempplate, positions, side_locs=side_locs, start_at=stage) as tp:
            tp.set_flow_rate(aspirate=self._elute_aspiration_rate, dispense=self._elute_dispense_rate)
            tp.pick_up()
            tp.aspirate(self._elution_vol, self.water)
            tp.air_gap(self._elute_air_gap)
            tp.dispense(self._elute_air_gap, locationFrom="target", well_modifier="top(0)")
            tp.dispense(self._elution_vol, locationFrom="side_locs")
            if self._elute_mix_times > 0:
                tp.mix(self._elute_mix_times, self._elute_mix_vol, locationFrom="side_locs")
            # tp.touch_tip(v_offset=self._touch_tip_height)
            tp.air_gap(self._elute_air_gap)
            tp.drop_tip()

    def final_transfer(self):
        locs = []
        for i, (m, e) in enumerate(zip(self.mag_samples_m, self.pcr_samples_m)):
            side = -1 if i % 2 == 0 else 1
            locs.append(m.bottom(self._final_transfer_dw_bottom_height).move(Point(x=side * self._final_transfer_side)))

        with PairedPipette(self._magplate, locs, start_at="final transfer", pcr_locs=self.pcr_samples_m) as tp:
            tp.set_flow_rate(aspirate=self._final_transfer_rate_aspirate, dispense=self._final_transfer_rate_dispense)
            tp.pick_up()
            tp.aspirate(self._final_vol, locationFrom="target")
            tp.air_gap(self._elute_air_gap)
            tp.dispense(self._elute_air_gap, locationFrom="pcr_locs", well_modifier="top(0)")
            tp.dispense(self._final_vol,
                        locationFrom="pcr_locs",
                        well_modifier="bottom({})".format(self._final_mix_height))
            tp.mix(self._final_mix_times, self._final_mix_vol, locationFrom="pcr_locs",
                   well_modifier="bottom({})".format(self._final_mix_height))
            tp.blow_out(locationFrom="pcr_locs",
                        well_modifier="top({})".format(self._final_mix_blow_out_height))
            tp.air_gap(self._elute_air_gap)
            tp.drop_tip()

    def wash(self, vol: float, source, mix_reps: int, wash_name: str = "wash"):
        self.logger.info(self.msg_format("wash info", vol, wash_name, mix_reps))
        if wash_name == "wash A":
            self._m300.flow_rate.dispense = self._wash_1_mix_dispense_rate
            self._m300r.flow_rate.dispense = self._wash_1_mix_dispense_rate
            self._m300.flow_rate.aspirate = self._wash_1_mix_aspiration_rate
            self._m300r.flow_rate.aspirate = self._wash_1_mix_aspiration_rate
            vertical_speed = self._wash_1_vertical_speed
        elif wash_name == "wash B":
            self._m300.flow_rate.dispense = self._wash_2_mix_dispense_rate
            self._m300r.flow_rate.dispense = self._wash_2_mix_dispense_rate
            self._m300.flow_rate.aspirate = self._wash_2_mix_aspiration_rate
            self._m300r.flow_rate.aspirate = self._wash_2_mix_aspiration_rate
            vertical_speed = self._wash_2_vertical_speed
        self._magdeck.disengage()

        num_trans, vol_per_trans = uniform_divide(vol, self._wash_max_transfer_vol)

        src = []
        mix_locs = []
        wash_locs = []
        wash_locs_index = [0, 0, 1, 1, 4, 4, 5, 5, 2, 2, 3, 3]

        for i, m in enumerate(self.mag_samples_m):
            wash_locs.append(source[wash_locs_index[i]])
            src.append(self.wash_getcol(i, len(self.mag_samples_m), source))
            mix_locs.append(m.bottom(self._bottom_headroom_height).move(Point(x=2 * (-1 if i % 2 else +1))))

        with PairedPipette(self._magplate, self.mag_samples_m,
                           wash_locs=wash_locs,
                           mix_locs=mix_locs,
                           start_at=wash_name) as tp:
            tp.pick_up()
            for n in range(num_trans):
                if n > 0:
                    tp.dispense(self._wash_air_gap, locationFrom="wash_locs", well_modifier="top(0)")

                tp.aspirate(vol_per_trans, locationFrom="wash_locs")
                tp.move_to(locationFrom="wash_locs", well_modifier="top(0)", speed=vertical_speed)
                tp.dispense(vol_per_trans, locationFrom="target", well_modifier="top(0)")
                if n < num_trans-1:
                    tp.air_gap(self._wash_air_gap)

            # Mix
            tp.mix(mix_reps, self._wash_mix_vol, locationFrom="mix_locs")
            tp.move_to(locationFrom="target", well_modifier="top(0)", speed=vertical_speed)
            tp.air_gap(self._wash_air_gap)
            tp.drop_tip()

        self._magdeck.engage(height=self._magheight)
        self.check()
        if self.run_stage("{} incubate".format(wash_name)):
            self.delay(self._wait_time_wash_on, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
        self.remove_supernatant(vol, stage="remove {}".format(wash_name))

if __name__ == "__main__":
    station = StationBTechnogeneticsPairedPipette(metadata={'apiLevel': '2.7'},
                                                  num_samples=96)
    # logging.getLogger(StationBTechnogeneticsPairedPipette.__name__).setLevel(logging.DEBUG)
    station.simulate()




import logging

from .utils import WellWithVolume, MoveWithSpeed


class MultiTubeSource(object):
    """

    """
    def __init__(self, name="", logger=None,
                 vertical_speed: int = None,
                 vertical_slow_start_overheight: int = 10):
        self._source_tubes_and_vol = []
        self._name = name
        self._vertical_speed = vertical_speed
        self._vertical_slow_start_overheight = vertical_slow_start_overheight
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)
        self._aspirate_list = []

    def append_tube_with_vol(self, source, available_volume):
        self._source_tubes_and_vol.append(dict(source=source,
                                               available_volume=available_volume))
        self.logger.debug("{}: appended {} with {}ul".format(self._name, source, available_volume))
        self.logger.debug("Now sources is: {}".format(self._source_tubes_and_vol))

    def prepare_aspiration(self, volume, fixed_height: float = None, min_height: float = 0.5):
        """ Prepare the aspiration list for the requested volume
        @parameter volume: the volume in ul to aspirate;
        @parameter fixed_height: if set force the aspiration at this height from bottom;
        @parameter min_height: the minimum height to reach from bottom with automatic height calculation"""
        self._aspirate_list = []
        left_volume = volume
        for source_and_vol in self._source_tubes_and_vol:
            if source_and_vol["available_volume"] >= left_volume:
                # aspirate_list.append(dict(source=source_and_vol["source"], vol=left_volume))
                aspirate_vol = left_volume
            else:
                aspirate_vol = source_and_vol["available_volume"]
            left_volume -= aspirate_vol
            source_and_vol["available_volume"] -= aspirate_vol
            if aspirate_vol != 0:
                if fixed_height is None:
                    height = WellWithVolume(
                        well=source_and_vol["source"],
                        initial_vol=source_and_vol["available_volume"],
                        min_height=min_height).height
                else:
                    height = fixed_height
                self._aspirate_list.append(dict(source=source_and_vol["source"], vol=aspirate_vol, height=height))

            if left_volume == 0:
                break
        else:
            raise Exception("{}: no volume left in source tubes.".format(self._name))

        self.logger.debug("{} sources: {}".format(self._name, self._source_tubes_and_vol))

    def aspirate(self, pip):
        assert self._aspirate_list, "You must call calculate_aspirate_volume before aspirate"
        for a in self._aspirate_list:
            if self._vertical_speed is not None:
                with MoveWithSpeed(pip,
                                   from_point=a["source"].bottom(a["height"] + self._vertical_slow_start_overheight),
                                   to_point=a["source"].bottom(a["height"]),
                                   speed=self._vertical_speed, move_close=False):
                    pip.aspirate(a["vol"])
            else:
                pip.aspirate(a["vol"], a["source"].bottom(a["height"]))

    @property
    def locations_str(self):
        return "{}: ".format(self._name) + \
               " ".join(["{}; ".format(sv['source'])
                         for sv in self._source_tubes_and_vol])

    def __str__(self):
        return "{}: ".format(self._name) +\
               " ".join(["{} with volume {}ul;".format(sv['source'], sv['available_volume'])
                        for sv in self._source_tubes_and_vol])



from opentrons.types import Mount, Point
from opentrons.hardware_control import CriticalPoint


class HardwareHelper:
    def __init__(self, ctx):
        self._ctx = ctx

    def get_gantry_position(self, m: Mount, update_position: bool = False) -> Point:
        return self._ctx._hw_manager.hardware.gantry_position(m, CriticalPoint.MOUNT, update_position)

    def move_rel(self, mount, delta: Point):
        return self._ctx._hw_manager.hardware.move_rel(mount, delta)

    def retract(self, mount):
        self._ctx._hw_manager.hardware.retract(mount)



from ..station import Station, labware_loader, instrument_loader
from ..utils import uniform_divide, get_labware_json_from_filename, MoveWithSpeed, WellWithVolume
from ..a.copan_48 import copan_48_corrected_specs


class StationABioerPreparation(Station):
    def __init__(self,
                 num_samples: int = 16,
                 deepwell_slot: int = 1,
                 deepwel_initial_volume: float = 600,
                 pk_tube_bottom_height: float = 2,
                 pk_dw_bottom_height: float = 13.5,
                 pk_rate_aspirate: float = 100,
                 pk_rate_dispense: float = 100,
                 pk_volume: float = 10,
                 pk_dead_volume: float = 5,
                 pk_vertical_speed: float = 40,
                 sample_vertical_speed: float = 40,
                 sample_air_gap: float = 25,
                 sample_rate_aspirate: float = 200,
                 sample_rate_dispense: float = 200,
                 sample_volume: float = 300,
                 sample_destination_bottom_height: float = 13.5,
                 source_racks_slots=[2],
                 source_bottom_height_aspirate: float = 6,
                 source_bottom_height_start: float = 40,
                 tube_block_model: str = 'opentrons_24_tuberack_nest_1.5ml_screwcap',
                 tube_rack_slot: int = 6,
                 p300_max_volume: float = 200,
                 ** kwargs):

        assert num_samples <= 16, "Protocol must be used with 16 or less samples"

        super(StationABioerPreparation, self).__init__(
            num_samples=num_samples,
            ** kwargs)
        self._pk_volume = pk_volume
        self._pk_rate_aspirate = pk_rate_aspirate
        self._pk_rate_dispense = pk_rate_dispense
        self._pk_vertical_speed = pk_vertical_speed
        self._sample_vertical_speed = sample_vertical_speed
        self._pk_tube_bottom_height = pk_tube_bottom_height
        self._pk_dw_bottom_height = pk_dw_bottom_height
        self._pk_dead_volume = pk_dead_volume
        self._deepwell_slot = deepwell_slot
        self._deepwel_initial_volume = deepwel_initial_volume
        self._sample_air_gap = sample_air_gap
        self._sample_rate_aspirate = sample_rate_aspirate
        self._sample_rate_dispense = sample_rate_dispense
        self._sample_volume = sample_volume
        self._sample_destination_bottom_height = sample_destination_bottom_height
        self._source_racks_slots = source_racks_slots
        self._source_bottom_height_start = source_bottom_height_start
        self._source_bottom_height_aspirate = source_bottom_height_aspirate
        self._tube_block_model = tube_block_model
        self._tube_rack_slot = tube_rack_slot
        self._p300_max_volume = p300_max_volume
        self._p1000_fake_aspirate = True


    @labware_loader(0, "_tips1000")
    def load_tips1000(self):
        self._tipracks_main = [self._ctx.load_labware('opentrons_96_filtertiprack_1000ul', slot, '1000µl filter tiprack')
                               for slot in ['10', '11']]

    @labware_loader(1, "_tipracks20")
    def load_tipracks20(self):
        self._tipracks20 = [self._ctx.load_labware('opentrons_96_filtertiprack_20ul', slot, '20ul filter tiprack')
                                for slot in ['8', '9']]

    @labware_loader(2, "_source_racks")
    def load_source_racks(self):
        labware_def = copan_48_corrected_specs.labware_definition()
        self._source_racks = [
            self._ctx.load_labware_from_definition(
                labware_def, slot,
                'source tuberack ' + str(i + 1)
            ) for i, slot in enumerate(self._source_racks_slots)
        ]

    @labware_loader(5, "_dests_plates")
    def load_dests_plate(self):
        self._dests_plates = [self._ctx.load_labware_from_definition(get_labware_json_from_filename("bioer_96_wellplate_2000ul.json"),
                                         slot,
                                         '96-deepwell sample plate {}'.format(i+1)) for i, slot in enumerate([self._deepwell_slot])]

    @instrument_loader(0, "_p1000")
    def load_p1000(self):
        self._p1000 = self._ctx.load_instrument('p1000_single_gen2', 'right', tip_racks=self._tipracks_main)

    @instrument_loader(0, "_m20")
    def load_m20(self):
        self._m20 = self._ctx.load_instrument('p20_multi_gen2', 'left', tip_racks=self._tipracks20)

    def _tipracks(self) -> dict:
         return {
            "_tipracks_main": "_p1000",
            "_tipracks20": "_m20",
        }

    @property
    def samples_destination_rows(self):
        return [row for plate in self._dests_plates for row in plate.columns()[::6]]

    @property
    def samples_destination_single(self):
        return [sample for row in self.samples_destination_rows for sample in row][:self._num_samples]

    @property
    def samples_source_single(self):
        return [sample for rack in self._source_racks for sample in rack.wells()]

    @property
    def proteinase_destinations(self):
        return self.samples_destination_single

    @property
    def p300_available_vol(self):
        return self._p300_max_volume - self._pk_dead_volume


    def transfer_sample(self, source, dest):
        self.logger.debug("transferring from {} to {}".format(source, dest))
        dest_with_volume = WellWithVolume(dest, self._deepwel_initial_volume, headroom_height=0)
        dest_initial_height = dest_with_volume.height

        self.pick_up(self._p1000)

        if self._p1000_fake_aspirate:
            self._p1000_fake_aspirate = False
            self._p1000.aspirate(self._sample_air_gap, source.top())
            self._p1000.dispense(self._sample_air_gap, source.top())

        # Aspirating from source tube
        with MoveWithSpeed(self._p1000,
                           from_point=source.bottom(self._source_bottom_height_start),
                           to_point=source.bottom(self._source_bottom_height_aspirate),
                           speed=self._sample_vertical_speed, move_close=False):
            self._p1000.aspirate(self._sample_volume)
        self._p1000.air_gap(self._sample_air_gap)

        # Dispensing at the right height
        dest_with_volume.fill(self._sample_volume)
        dest_filled_height = dest_with_volume.height + 10

        with MoveWithSpeed(self._p1000,
                           from_point=dest.bottom(dest_filled_height),
                           to_point=dest.bottom(dest_initial_height),
                           speed=self._sample_vertical_speed, move_close=False):
            self._p1000.dispense(self._sample_volume + self._sample_air_gap)

        self._p1000.air_gap(self._sample_air_gap)
        self.drop(self._p1000)

    def transfer_samples(self, stage: str = "transfer samples"):
        self._p1000.flow_rate.aspirate = self._sample_rate_aspirate
        self._p1000.flow_rate.dispense = self._sample_rate_dispense

        sources_and_dest_list = list(zip(self.samples_source_single, self.samples_destination_single))
        for done_count, (s, d) in enumerate(sources_and_dest_list):
            if self.run_stage("{} {}/{}".format(stage, done_count+1, len(sources_and_dest_list))):
                self.transfer_sample(s, d)

    def body(self):
        self.transfer_samples(stage="transfer samples")

# protocol for loading in Opentrons App or opentrons_simulate
# =====================================
#logging.getLogger(BioerPreparation.__name__).setLevel(logging.INFO)

metadata = {'apiLevel': '2.7'}
station = StationABioerPreparation(num_samples = 16)

def run(ctx):
    return station.run(ctx)

if __name__ == "__main__":
    StationABioerPreparation(metadata={'apiLevel':'2.7'}).simulate()



from .b import StationB, labware_loader
from typing import Tuple
from opentrons.types import Point
from ..utils import get_labware_json_from_filename, mix_bottom_top, WellWithVolume


class StationBTechnogenetics(StationB):
    _protocol_description = "station B protocol for Technogenetics kit"

    def __init__(self,
                 beads_drying_time: float = 5,
                 elute_mix_times: int = 0,
                 elution_vol: float = 50,
                 elute_incubate: bool = False,
                 final_mix_height: float = 0.5,
                 final_mix_times: int = 5,
                 final_mix_vol: float = 20,
                 final_mix_blow_out_height: float = -2,
                 final_transfer_rate_aspirate: float = 30,
                 final_transfer_rate_dispense: float = 30,
                 final_transfer_side: float = 0.5,
                 final_transfer_dw_bottom_height: float = 0.6,
                 final_vol: float = 20,
                 flatplate_slot: str = '3',
                 incubation_temperature: float = 55,
                 incubation_time: float = 20,
                 h_bottom: float = 1,
                 n_bottom: float = 3,
                 mix_incubate_on_time: float = 20,
                 mix_incubate_off_time: float = 5,
                 postspin_incubation_time: float = 3,
                 remove_wash_vol: float = 50,
                 sample_mix_height: float = 0.3,
                 sample_mix_times: int = 10,
                 sample_mix_vol: float = 180,
                 sample_vertical_speed: float = 35,
                 mix_samples_rate = 200,
                 mix_samples_last_rate = 94,
                 starting_vol: float = 650,
                 supernatant_removal_aspiration_rate_first_phase = 94,
                 tempdeck_slot: str = '10',
                 tempdeck_temp: float = None,
                 tempdeck_auto_turnon: bool = False,
                 tempdeck_auto_turnoff: bool = True,
                 thermomixer_incubation_time: float = 5,
                 tipracks_slots: Tuple[str, ...] = ('4', '6', '7', '8', '9'),
                 wash_1_vol: float = 650,
                 wash_2_vol: float = 650,
                 wash_headroom: float = 1.05,
                 watchdog_serial_timeout_seconds: int = 30,
                 **kwargs
                 ):
        """ Build a :py:class:`.StationBTechnogenetics`.
        :param beads_drying_time: [minutes] time to wait for beads to try in air
        :param external_deepwell_incubation: whether or not to perform deepwell incubation outside the robot
        :param final_mix_height: Mixing height (from the bottom) for final transfer in mm
        :param final_mix_times: Mixing repetitions for final transfer
        :param final_mix_vol: Mixing volume for final transfer in uL
        :param final_transfer_rate_aspirate: Aspiration rate during final transfer in uL/s
        :param final_transfer_rate_dispense: Dispensation rate during final transfer in uL/s
        :param final_vol: Volume to transfer to the PCR plate in uL
        :param incubation_temperature: Temperature set on temperature module to incubate samples in °C
        :param incubation_time: minutes to wait on temperature module to incubate samples
        :param mix_incubate_on_time: Time for incubation on magnet after mix in minutes
        :param mix_incubate_off_time: Time for incubation off magnet after mix in minutes
        :param postspin_incubation_time: Post-spin incubation time in minutes
        :param remove_wash_vol: Volume to remove during wash removal in uL
        :param sample_mix_times: Mixing height for samples in mm from the bottom
        :param sample_mix_times: Mixing repetitions for samples
        :param sample_mix_vol: Mixing volume for samples in uL
        :param sample_vertical_speed: speed of vertical movement exiting well
        :param thermomixer_incubation_time: Time for incubation after thermomixer in minutes
        """
        super(StationBTechnogenetics, self).__init__(
            elute_mix_times=elute_mix_times,
            elution_vol=elution_vol,
            elute_incubate=elute_incubate,
            flatplate_slot=flatplate_slot,
            starting_vol=starting_vol,
            supernatant_removal_aspiration_rate_first_phase = supernatant_removal_aspiration_rate_first_phase,
            tempdeck_slot=tempdeck_slot,
            tempdeck_temp=tempdeck_temp,
            tempdeck_auto_turnon=tempdeck_auto_turnon,
            tempdeck_auto_turnoff=tempdeck_auto_turnoff,
            tipracks_slots=tipracks_slots,
            wash_1_vol=wash_1_vol,
            wash_2_vol=wash_2_vol,
            wash_headroom=wash_headroom,
            **kwargs
        )
        self._beads_drying_time = beads_drying_time
        self._final_mix_blow_out_height = final_mix_blow_out_height
        self._final_mix_height = final_mix_height
        self._final_mix_times = final_mix_times
        self._final_transfer_side = final_transfer_side
        self._final_transfer_dw_bottom_height = final_transfer_dw_bottom_height
        self._final_mix_vol = final_mix_vol
        self._final_transfer_rate_aspirate = final_transfer_rate_aspirate
        self._final_transfer_rate_dispense = final_transfer_rate_dispense
        self._mix_samples_rate = mix_samples_rate
        self._mix_samples_last_rate = mix_samples_last_rate
        self._final_vol = final_vol
        self._flatplate_slot = flatplate_slot
        self._incubation_temperature = incubation_temperature
        self._incubation_time = incubation_time
        self._h_bottom = h_bottom
        self._n_bottom = n_bottom
        self._mix_incubate_on_time = mix_incubate_on_time
        self._mix_incubate_off_time = mix_incubate_off_time
        self._postspin_incubation_time = postspin_incubation_time
        self._remove_wash_vol = remove_wash_vol
        self._sample_mix_height = sample_mix_height
        self._sample_mix_times = sample_mix_times
        self._sample_mix_vol = sample_mix_vol
        self._sample_vertical_speed = sample_vertical_speed
        self._thermomixer_incubation_time = thermomixer_incubation_time
        self._watchdog_serial_timeout_seconds = watchdog_serial_timeout_seconds
    
    @labware_loader(5, "_flatplate")
    def load_flatplate(self):
        self._flatplate = self._ctx.load_labware('opentrons_96_aluminumblock_nest_wellplate_100ul', self._flatplate_slot, 'chilled elution plate on block for Station C')
    
    @labware_loader(5, "_tempplate")
    def load_tempplate(self):
        self._tempplate = self._tempdeck.load_labware(self._magplate_model)

    @labware_loader(6, "_waste")
    def load_waste(self):
        self._waste = self._ctx.load_labware_from_definition(
            get_labware_json_from_filename("biofil_1_reservoir_500000ul.json"), '11', 'Liquid Waste').wells()[0].top()

    @property
    def pcr_samples_m(self):
        return self._flatplate.rows()[0][:self.num_cols]
    
    @property
    def temp_samples_m(self):
        return self._tempplate.rows()[0][:self.num_cols]
    
    def load_etoh(self): pass

    @labware_loader(9, "_elut12")
    def load_elut12(self):
        self._elut12 = self._ctx.load_labware('nest_12_reservoir_15ml', '2', 'Trough with Wash B and Elution')

    @property
    def water(self):
        return self._elut12.wells()[11]

    @property
    def wash1(self):
        return self._res12.wells()[:6]
    
    @property
    def wash2(self):
        return self._elut12.wells()[:6]

    
    @staticmethod
    def wash_getcol(sample_col_idx: int, wash_cols: int, source):
        return source[sample_col_idx // 2]
    
    def mix_samples(self):
        for i, m in enumerate(self.temp_samples_m):
            if self.run_stage("mix sample {}/{}".format(i + 1, len(self.mag_samples_m))):
                self._m300.flow_rate.aspirate = self._mix_samples_rate
                self._m300.flow_rate.dispense = self._mix_samples_rate

                well_with_volume = WellWithVolume(m,
                                                  initial_vol=self._starting_vol - self._sample_mix_vol,
                                                  min_height=self._sample_mix_height,
                                                  headroom_height=0)
                self.pick_up(self._m300)
                mix_bottom_top(pip=self._m300,
                               reps=self._sample_mix_times,
                               vol=self._sample_mix_vol,
                               pos=m.bottom,
                               bottom=self._sample_mix_height,
                               top=well_with_volume.height,
                               last_dispense_rate=self._mix_samples_last_rate)

                self._m300.move_to(m.top(0), speed=self._sample_vertical_speed)
                self._m300.air_gap(self._bind_air_gap)
                self.drop(self._m300)
    
    def elute(self, positions=None, transfer: bool = False, stage: str = "elute"):
        if positions is None:
            positions = self.temp_samples_m
        self._magdeck.disengage()
        super(StationBTechnogenetics, self).elute(positions=positions, transfer=transfer, stage=stage)
        self._magdeck.disengage()
    
    def remove_wash(self, vol, stage: str = "remove wash"):
        self.remove_supernatant(vol, stage)
    
    def final_transfer(self):
        self._m300.flow_rate.aspirate = self._final_transfer_rate_aspirate
        self._m300.flow_rate.dispense = self._final_transfer_rate_dispense
        n = len(list(zip(self.mag_samples_m, self.pcr_samples_m)))

        for i, (m, e) in enumerate(zip(self.mag_samples_m, self.pcr_samples_m)):
            if self.run_stage("final transfer {}/{}".format(i + 1, n)):
                self.pick_up(self._m300)
                side = -1 if i % 2 == 0 else 1
                loc = m.bottom(self._final_transfer_dw_bottom_height).move(Point(x=side*self._final_transfer_side))
                self._m300.aspirate(self._final_vol, loc)
                self._m300.air_gap(self._elute_air_gap)
                self._m300.dispense(self._elute_air_gap, e.top())
                self._m300.dispense(self._m300.current_volume, e.bottom(self._final_mix_height))

                #self._m300.transfer(self._final_vol, loc, e.bottom(self._elution_height), air_gap=self._elute_air_gap, new_tip='never')
                self._m300.mix(self._final_mix_times, self._final_mix_vol, e.bottom(self._final_mix_height))
                self._m300.blow_out(e.top(self._final_mix_blow_out_height))
                self._m300.air_gap(self._elute_air_gap)
                self.drop(self._m300)

    def body(self):
        self.logger.info(self.get_msg_format("volume", "wash 1", self._wash_headroom * self._wash_1_vol * self._num_samples / 1000))
        self.logger.info(self.get_msg_format("volume", "wash 2", self._wash_headroom * self._wash_2_vol * self._num_samples / 1000))
        self.logger.info(self.get_msg_format("volume", "elution buffer", self._wash_headroom * self._elution_vol * self._num_samples / 1000))

        self.tempdeck_set_temperature(self._incubation_temperature)

        self.delay_start_count()
        self.mix_samples()
        if self.run_stage("incubation"):
            self.delay_wait_to_elapse(minutes=self._incubation_time)

        self.tempdeck_deactivate()

        self.pause("move plate to magdeck")

        self._magdeck.engage(height=self._magheight)
        self.check()
        if self.run_stage("mix incubate off"):
            self.delay(self._mix_incubate_off_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))

        self.remove_supernatant(self._starting_vol)
        self.wash(self._wash_1_vol, self.wash1, self._wash_1_times, "wash A")

        self._magdeck.disengage()

        if self.run_stage("spin deepwell wash A"):
            self.dual_pause("spin the deepwell", between=self.set_external)
            self.set_internal()
            self.dual_pause("Add wash B and elute buffer in slot {}{}".format(
                self.wash2[0].parent, " and {}".format(self.water.parent) if self.wash2[0].parent != self.water.parent else ""))

        self._magdeck.engage(height=self._magheight)
        self.check()

        if self.run_stage("post spin incubation wash A"):
            self.delay(self._postspin_incubation_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))

        self.remove_wash(self._remove_wash_vol, "remove wash A after spin")

        self._magdeck.disengage()

        self.wash(self._wash_2_vol, self.wash2, self._wash_2_times, "wash B")

        self._magdeck.disengage()
        if self.run_stage("spin deepwell wash B"):
            self.dual_pause("spin the deepwell", between=self.set_external)
            self.set_internal()

        self._magdeck.engage(height=self._magheight)
        self.check()

        if self.run_stage("post spin incubation wash B"):
            self.delay(self._postspin_incubation_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))

        if self._tempdeck_temp is not None and not self._tempdeck_auto_turnon:
            # self._tempdeck.start_set_temperature(self._tempdeck_temp)
            self.tempdeck_set_temperature(self._tempdeck_temp)

        self.remove_wash(self._remove_wash_vol, "remove wash B after spin")

        self._magdeck.disengage()

        if self.run_stage("beads drying"):
            self.delay(self._beads_drying_time, msg="beads drying")

        self.elute(self.mag_samples_m)

        if self.run_stage("thermomixer"):
            self.dual_pause("seal the deepwell", between=self.set_external)
            self.set_internal()

        self._magdeck.engage(height=self._magheight)
        self.check()
        if self.run_stage("post thermomixer incubation"):
            self.delay(self._thermomixer_incubation_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))

        if self.run_stage("input PCR"):
            self.dual_pause("input PCR")

        self.final_transfer()

        self._magdeck.disengage()
        self.logger.info(self.msg_format("move to PCR"))

    def tempdeck_set_temperature(self, temperature):
        self.watchdog_reset(self._watchdog_serial_timeout_seconds)
        self._tempdeck.start_set_temperature(temperature)
        self.watchdog_stop()

    def tempdeck_deactivate(self):
        self.watchdog_reset(self._watchdog_serial_timeout_seconds)
        self._tempdeck.deactivate()
        self.watchdog_stop()

if __name__ == "__main__":
    StationBTechnogenetics(metadata={'apiLevel': '2.3'}).simulate()


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



# -------------------------------------------------------------------------
from opentrons.protocols.api_support.util import labware_column_shift
from opentrons.protocol_api.paired_instrument_context import PairedInstrumentContext, PipettePairPickUpTipError
from opentrons.protocol_api.labware import Well, Labware
from opentrons import types
import logging
import parse
from opentrons.types import Mount, Point, Location


class PairedPipette:
    available_commands = ["pick_up", "drop_tip", "mix", "air_gap", "aspirate", "dispense", "move_to", "touch_tip",
                          "comment", "set_flow_rate", "blow_out"]
    pips = []
    pippairedctx = None
    labware_height_overhead = 10.0      # mm height over the top of the tallest labware

    @classmethod
    def setup(cls, pipette1, pipette2, stationctx, pick_up_single: bool = False):
        cls.pips = [pipette1, pipette2]
        cls.pippairedctx = pipette1.pair_with(pipette2)
        cls._stationctx = stationctx
        cls._pick_up_single = pick_up_single
        cls.max_labware_height = cls.labware_height_overhead + \
            max([labware.highest_z for labware in cls._stationctx._ctx.loaded_labwares.values()])

    def __init__(self, labware,
                 targets,
                 start_at: str = None,
                 **kwargs):
        # dests è un iterabile che contiene le destinazioni
        assert isinstance(labware, Labware), "Paired pipette labware not a labware but a {}".format(type(labware))
        self.labware = labware
        self.donedests = []
        self.switchpipette = 1  ## used to switch between pipettes when single pipette is requested
        self.commands = []
        self._logger = logging.getLogger(PairedPipette.__name__)
        self._logger.setLevel(logging.INFO)
        self._locations = {'target': targets}
        self._locations_as_well = [self._get_well_from_location(location) for location in self._locations['target']]
        for kwarg in kwargs:
            self._logger.debug("Appending {} to keywork {}".format(kwargs[kwarg], kwarg))
            self._locations[kwarg] = kwargs[kwarg]
        self._logger.debug("Locations now contain: {}".format(self._locations))
        self._start_at = start_at
        if self._start_at:
            assert callable(getattr(self._stationctx, "run_stage")), \
                "run_stage function not found. Please do not use start_at names."
            self._start_at_function = getattr(self._stationctx, "run_stage")
            self._start_at += " {}/{}"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._run()

    def getctx(self, dest_well):
        assert self.__class__.pips, "You must initialize the module with setup first"
        try:
            _, secondary_well = self.pippairedctx._get_locations(dest_well)
            self._logger.debug("Primary well: {}. Secondary well found: {}".format(dest_well, secondary_well))
            if secondary_well in self._locations_as_well and secondary_well not in self.donedests:
                self._logger.debug("We can use paired pipette on destination.")
                return self.__class__.pippairedctx, secondary_well
        except PipettePairPickUpTipError as e:
            self._logger.debug("We can't use paired pipette here. Exception: {}".format(e))
        pipette_to_use = self._get_single_pipette()
        self._check_pipette_over_labware(self.__class__._get_other_pipette(pipette_to_use))
        return pipette_to_use, None

    def _get_single_pipette(self):
        """ Get a single pipette to use for a single-pipette operation"""
        # self.switchpipette = (self.switchpipette + 1) % len(self.__class__.pips)
        # return self.__class__.pips[self.switchpipette]
        # for now use only one pipette if single pipette needed.
        # TODO maybe switch pipette each run?
        return self.__class__.pips[0]

    def _pick_up_tip(self, pipctx):
        if isinstance(pipctx, PairedInstrumentContext):
            if self.__class__._pick_up_single:
                self.pick_up_tip_single()
            else:
                self.pick_up_tip_paired(pipctx)
        else:
            self._stationctx.pick_up(pipctx)

    def pick_up_tip_single(self):
        for p in self.pips:
            self._stationctx.pick_up(p)

    def pick_up_tip_paired(self, pipctx):
        try:
            self._stationctx.pick_up(pipctx)
        except PipettePairPickUpTipError:
            self._logger.debug("Paired pipette needed but cannot pickup tips, doing one pipette at a time")
            self.pick_up_tip_single()

    def _drop_tip(self, pipctx):
        try:
            self._stationctx.drop(pipctx)
        except TypeError:
            if isinstance(pipctx, PairedInstrumentContext):
                self._logger.debug("Pip ctx {}: probably pick_up_tip has been done with single pipette".format(pipctx))
                self._check_pipette_over_labware(pipctx)    # before move single pipette bring each pipette in safe position
                for p in self.pips:
                    self._stationctx.drop(p)

    def _check_pipette_over_labware(self, pipctx):
        """Check that the pipette has enough height to pass over all labware"""
        mount = pipctx._pair_policy.primary if isinstance(pipctx,
                                                          PairedInstrumentContext) else pipctx._implementation.get_mount()
        current_loc = self._stationctx._ctx._hw_manager.hardware.gantry_position(mount)
        self._logger.debug("{}: Actual location for mount {} is: {}".format(pipctx, mount, current_loc))
        if current_loc.z < self.__class__.max_labware_height:
            new_point = Point(x=current_loc.x, y=current_loc.y, z=self.__class__.max_labware_height)
            self._logger.debug("Location to move to: {}".format(new_point))
            pipctx.move_to(Location(new_point, None))

    @classmethod
    def _set_flow_rate(cls, aspirate: str, dispense: str):
        for p in cls.pips:
            if aspirate:
                p.flow_rate.aspirate = aspirate
            if dispense:
                p.flow_rate.dispense = dispense

    @staticmethod
    def substitute_kwarg_location(new_location, keyword, kwargs):
        if keyword in kwargs:
            call, arguments = parse.parse("{}({})", kwargs[keyword])
            # arguments conversion to string
            if call in ['top', 'bottom']:
                arguments = float(arguments)

            kwargs.pop(keyword)
            kwargs['location'] = getattr(new_location, call)(arguments)

    def substitute_kwargs_locations(self, index, kwargs):
        if 'locationFrom' in kwargs:
            assert isinstance(kwargs['locationFrom'], str), "location parameter must be a string."
            if kwargs['locationFrom'] in self._locations:
                loc_keyword = kwargs.pop('locationFrom')
                assert 'location' not in kwargs, "location already set, not overwriting. Value: {}".format(kwargs['location'])
                kwargs['location'] = self._locations[loc_keyword][index]
            else:
                raise Exception("Location {} not set.".format(kwargs['locationFrom']))

    def substitute_kwargs_well_modifier(self, kwargs):
        if 'well_modifier' in kwargs:
            if 'location' in kwargs:
                self._logger.debug("Substitute kwargs {}".format(kwargs['location']))
                call, arguments = parse.parse("{}({})", kwargs['well_modifier'])
                # arguments conversion to string
                if call in ['top', 'bottom']:
                    arguments = float(arguments)
                kwargs.pop('well_modifier')
                old_location = kwargs.pop('location')
                kwargs['location'] = getattr(old_location, call)(arguments)

    def substitute_kwargs_move_points(self, kwargs, index):
        """ input: kwargs - arguments that in 'well_move_points' contains an array of point;
                   each point is applied to the target with a *move* instruction.
            input: index -  actual index of target location
        """
        if 'well_move_points' in kwargs:
            if len(kwargs['well_move_points']) <= index:
                raise Exception("Move point array must be as long as or longer than target array.")
            if 'location' in kwargs:
                self._logger.debug("Substitute kwargs {}".format(kwargs['location']))
                points_array = kwargs.pop('well_move_points')
                old_location = kwargs.pop('location')
                kwargs['location'] = old_location.move(points_array[index])

    @classmethod
    def _get_other_pipette(cls, pip):
        for p in cls.pips:
            if p != pip:
                return p
        raise Exception("Different pipette not found")

    @staticmethod
    def _get_well_from_location(loc) -> Well:
        if loc and isinstance(loc, types.Location):
            if loc.labware.is_well:
                well = loc.labware.as_well()
            else:
                raise Exception("Cannot make Well from destination {} of type {}.".format(loc, type(loc)))
        elif loc and isinstance(loc, Well):
            well = loc
        else:
            raise Exception("Destination {} of type {} not recognized.".format(loc, type(loc)))
        return well

    def _run(self):
        for i, (d, d_well) in enumerate(zip(self._locations['target'], self._locations_as_well)):
            self._logger.debug("Well is: {}".format(d_well))
            if d_well not in self.donedests:
                pipctx, secondary_well = self.getctx(d_well)
                if self._start_at and self._start_at_function(self._start_at.format(i+1, len(self._locations['target']))):
                    self._execute_command_list_on(d_well, secondary_well, pipctx)
                else:
                    self._logger.debug("dest {} skipped, not reached start_at".format(d_well))
                self.donedests.append(d_well)
                if secondary_well is not None:
                    self.donedests.append(secondary_well)
                self._logger.debug("donedests contains: {}".format(self.donedests))
            else:
                self._logger.debug("dest {} skipped, already done.".format(d_well))




    @staticmethod
    def _getValueFromKwargsAndClean(kwargs, key: str, default_value):
        if key in kwargs:
            ret_value = kwargs.pop(key)
        else:
            ret_value = default_value
        return ret_value

    def _execute_command_list_on(self, well, secondary_well, pipctx):
        primary_well_index = self._locations_as_well.index(well)
        secondary_well_index = self._locations_as_well.index(secondary_well) if secondary_well else None
        self._logger.debug("Using pipette context: {}".format(pipctx))
        skip_next_command = False       # skip_next_command is for merging aspirate and air-gap on each pipette
        for j, c in enumerate(self.commands):
            self._logger.debug("Evaluating command: {}".format(c))
            if skip_next_command:
                self._logger.debug("Asked to skip command.")
                skip_next_command = False
                continue

            if c['command'] == "pick_up":
                self._pick_up_tip(pipctx)
            elif c['command'] == "drop_tip":
                self._drop_tip(pipctx)
            elif c['command'] == "comment":
                self._stationctx._ctx.comment(*c['args'], **c['kwargs'])
            elif c['command'] == "set_flow_rate":
                self._set_flow_rate(*c['args'], **c['kwargs'])
            else:
                substituted_kwargs = dict(c['kwargs'])  # copy kwargs and substitute time by time
                # substituting kwargs to represent actual data
                is_target_paired = 'locationFrom' in substituted_kwargs
                is_forced_single = self._getValueFromKwargsAndClean(substituted_kwargs, 'forceSingle', False)
                is_command_ok_with_paired = self._getValueFromKwargsAndClean(substituted_kwargs, 'isOkWithPaired', False)
                self._logger.debug("Target is paired: {}; forced single: {}; command ok with paired: {}"
                                   .format(is_target_paired, is_forced_single, is_command_ok_with_paired))
                self.substitute_kwargs_locations(primary_well_index, substituted_kwargs)
                self.substitute_kwargs_well_modifier(substituted_kwargs)
                self.substitute_kwargs_move_points(substituted_kwargs, primary_well_index)

                to_do_with_single =  (is_target_paired==False and is_command_ok_with_paired==False) or is_forced_single
                self._logger.debug("To do with single is: {}".format(to_do_with_single))

                if isinstance(pipctx, PairedInstrumentContext) and to_do_with_single:
                    self._logger.debug("Using single pipetting")
                    substituted_kwargs_pip2 = dict(c['kwargs'])
                    self._getValueFromKwargsAndClean(substituted_kwargs_pip2, 'forceSingle', False)
                    self._getValueFromKwargsAndClean(substituted_kwargs_pip2, 'isOkWithPaired', False)
                    self.substitute_kwargs_locations(secondary_well_index, substituted_kwargs_pip2)
                    self.substitute_kwargs_well_modifier(substituted_kwargs_pip2)
                    self.substitute_kwargs_move_points(substituted_kwargs_pip2, secondary_well_index)

                    for (p, kwargs) in zip(self.__class__.pips, [substituted_kwargs, substituted_kwargs_pip2]):
                        self._logger.debug("Pipette {} command: {} args: {} {}".format(p, c['command'], c['args'], kwargs))
                        self._check_pipette_over_labware(self.__class__._get_other_pipette(p))
                        getattr(p, c['command'])(*c['args'], **kwargs)
                        # Merging air_gap with preceeding command
                        if len(self.commands) > (j + 1) and self.commands[j + 1]['command'] == 'air_gap':
                            self._logger.debug("Doing air_gap on {}".format(p))
                            self._getValueFromKwargsAndClean(self.commands[j + 1]['kwargs'], 'forceSingle', False)
                            self._getValueFromKwargsAndClean(self.commands[j + 1]['kwargs'], 'isOkWithPaired', False)
                            getattr(p, 'air_gap')(*self.commands[j + 1]['args'],
                                                  **self.commands[j + 1]['kwargs'])
                            skip_next_command = True
                else:
                    self._logger.debug("Pipette {} command: {} args: {} {}".format(pipctx, c['command'], c['args'],
                                                                      substituted_kwargs))
                    getattr(pipctx, c['command'])(*c['args'], **substituted_kwargs)
        return secondary_well


    def setcommand(self, command, *args, **kwargs):
        assert command in self.available_commands, "Command {} not recognized.".format(command)
        self.commands.append({'command': command, 'args': list(args), 'kwargs': kwargs})

    # API
    # as a real pipette apart from "source" and "dest" location keyword
    #
    # keep in mind that can be used only one location at a time (eg. no transfer from source to dest):
    # - source keywork (eg. source="bottom(1)") will be replaced with location=s.bottom(1)
    # - dest keywork (eg. dest="top(-2)") will be replaced with location=d.top(-2)

    def pick_up(self):
        self.setcommand('pick_up')

    def drop_tip(self):
        self.setcommand('drop_tip')

    def air_gap(self, *args, **kwargs):
        kwargs['isOkWithPaired'] = True
        self.setcommand('air_gap', *args, **kwargs)

    def mix(self, *args, **kwargs):
        self.setcommand('mix', *args, **kwargs)

    def aspirate(self, *args, **kwargs):
        self.setcommand('aspirate', *args, **kwargs)

    def dispense(self, *args, **kwargs):
        self.setcommand('dispense', *args, **kwargs)

    def move_to(self, *args, **kwargs):
        self.setcommand('move_to', *args, **kwargs)

    def touch_tip(self, *args, **kwargs):
        kwargs['isOkWithPaired'] = True
        self.setcommand('touch_tip', *args, **kwargs)

    def comment(self, *args, **kwargs):
        self.setcommand('comment', *args, **kwargs)

    def set_flow_rate(self, aspirate: str = None,
                            dispense: str = None):
        self.setcommand('set_flow_rate', aspirate=aspirate, dispense=dispense)

    def blow_out(self, *args, **kwargs):
        self.setcommand('blow_out', *args, **kwargs)




from .a import StationA
from ..station import StationMeta
import math


# Mixin allows for finer control over the mro
class StationAReloadMixin(metaclass=StationMeta):
    @property
    def max_samples_per_set(self) -> int:
        return len(self._sources)
    
    @property
    def sets_of_samples(self) -> int:
        return math.ceil(self._num_samples/self.max_samples_per_set)

    @property
    def remaining_samples(self) -> int:
        return self._num_samples - self._done_samples

    def transfer_samples(self):
        self._done_samples = 0
        refills = self.sets_of_samples - 1

        destinations = list(self.non_control_dests())

        self.logger.info(self.msg_format("refills", self.max_samples_per_set, refills))
        for set_idx in reversed(range(self.sets_of_samples)):
            self.logger.debug("{} remaining samples".format(self.remaining_samples))
            for s, d in zip(self._sources[:self.remaining_samples], self._dests_single[self._done_samples:]):
                if self.run_stage("transfer sample {}/{}".format(self._done_samples + 1, self._num_samples)):
                    if d in destinations:
                        self.transfer_sample(s, d)
                    else:
                        self._ctx.comment("Skipping transfer sample {}: is control".format(d))
                self._done_samples += 1
            if set_idx and self.run_stage("refill {}/{}".format(self.sets_of_samples - set_idx, self.sets_of_samples - 1)):
                self.request_dashboard_input(self.msg_format("refill", min(self.remaining_samples, self.max_samples_per_set)))

# Subclass is more straightforward
class StationAReload(StationAReloadMixin, StationA):
    pass


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
